#!/bin/perl
#################################################################################
# Created       : Martin Foster
# Modified      : 12-Jul-2003
#################################################################################
#
# Configure Puppet - Script part of Ethereal Realms, designed to configure 
#                    manage and remove puppets
# Copyright (C) 2000-2003  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 204, 817 5TH ST NE
#                 Calgary, Alberta
#                 T2E 3W9
#
#################################################################################

use CGI qw(:standard);						# Common gateway interface
use CGI::Carp qw(fatalsToBrowser);				# CGI Error Logs
use strict;							# Strict variable enforcement

use eRealms::Comm;						# Communications handler
use eRealms::Database;						# Database handler
use eRealms::Filter;						# Filter system
use eRealms::Home;						# Homepage/file handler
use eRealms::Image;						# Image handler
use eRealms::Login;						# Login functionality
use eRealms::Option;						# Options handling
use eRealms::Param;						# Parameter control
use eRealms::Table;						# Table control


#################################################################################
# Data Members
#################################################################################
my $cgi;							# Common gateway interface handle
my $database;							# Database handle
my $login;							# Login handle
my $param;							# Parameter handle
my $table;							# Table handle
my $tmpl;							# Template handle

my @are;							# Are so cool
my @not;							# Not so cool
my @act;							# Activated puppets
my @sus;							# Suspended puppets
my @targets;							# List of targets

my %config;							# Configuration tag

#################################################################################
# Program Area
#################################################################################

	# Initial handles
	$cgi      = new CGI;
	$database = new eRealms::Database();

	# Connect and fetch
	$database->Connect();

	# Pre-fetch active and suspended
	@are     = $cgi->param('ARE');
	@not     = $cgi->param('NOT');
	@act     = $cgi->param('APUPPET');
	@sus     = $cgi->param('IPUPPET');
	@targets = $cgi->param('TARGET');


	# Parameter handling
	$param = new eRealms::Param($database, $cgi);
	$param->GetParam();


	# Fetch hash
	$database->GetHashConfiguration(\%config);

	# Create instances
	$table = new eRealms::Table($cgi);
	$tmpl  = new eRealms::Template(\%config);


	# Document header
	print $cgi->header();
	$database->DocumentGetHeader();

	# Login handle
	$login = new eRealms::Login($database, $cgi, $param);

	# Authenthication
	if ($login->GetVerificationNormal())
	{


		# Puppet activity handling
		if (defined($param->{'ACTIVITY'}))
		{ 
			# Take care of business
			if (defined($param->{'MANAGE'}))
			{
				# Will activate suspend as necessary
				ActivityHandler($database, $cgi, $param, \%config, \@act, \@sus); 
			}

			# Display information
			ActivityDisplay($database, $cgi, $param, $table, $tmpl, \%config);
		}

		# Here comes Dolly
		elsif (defined($param->{'CLONE'}))
		{ PuppetClone($database, $cgi, $param, $table, $tmpl, \%config, \@targets); } 

		# Groupings and tasks
		elsif (defined($param->{'GROUP'}))
		{ PuppetGroup($database, $cgi, $param, $table, $tmpl, \%config, \@are, \@not); } 

		# Default behaviour
		else
		{
			# For additions and removal
			unless  ((defined($param->{$config{'TxtSubmitView'}})) 
			 || (defined($param->{$config{'TxtSubmitButton'}})))
			{ InitialDisplay($database, $cgi, $param, $table, $tmpl, \%config); }

			# For puppet changes and viewing
			else
			{ PuppetControl($database, $cgi, $param, $table, $tmpl, \%config); }
		}

	}

	# Document footer	
	$database->DocumentGetFooter();


#################################################################################
# Sub-Routines
#################################################################################

#####################
# Activity Display
#
# Handles the general output of the initial display screen.  Which will allow one
# to change password, E-Mail addresses and/or passwords.

sub ActivityDisplay
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table hander
	my $tmpl     = shift;					# Template handler
	my $config   = shift;					# Configuration hash

	my $url;						# Self referencing URL

	my $active;						# Active handles
	my $inactive;						# Inactive handles
	my $inline;						# Inline parameters

	my @active;						# Active handles
	my @inactive;						# Inactive handles
	
	my %args;						# Arguments
	

	#####################
	# Program area	
	

	# Back button
	# Generate Inline and URL
	$inline = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'}, ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'});	
	$url    = $cgi->url(-absolute=>1);

	
	# Puppet Ignore/Suspend
	# Populate arrays
	$database->GetListPuppetYours(\@active, $param->{'USER'});
	$database->GetListPuppetSuspended(\@inactive, $param->{'USER'});

	# Format lists
	$active   = $cgi->scrolling_list('APUPPET', \@active,   $active[0],   5, 'true');
	$inactive = $cgi->scrolling_list('IPUPPET', \@inactive, $inactive[0], 5, 'true');

	# Display of information
	# Title
	$tmpl->ShowMain($config->{'TagConfActivate'});

	# Form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), MANAGE=>'True');

	# Activation section
	$table->MakeTop();
	$table->MakeSingle("<CENTER>" 
	  . $tmpl->Pass('TagTabbed', MLINK=>"$url$inline", MNAME=>$config->{TxtBack})
	  . "</CENTER>"
	);
	$table->MakeBlank();

	$table->MakeSingle($tmpl->PassSub($config->{'TagActive'}));
	$table->MakeValid($active,   $cgi->submit($config->{'TxtInactive'}), 'middle');
	$table->MakeBlank();
	$table->MakeSingle($tmpl->PassSub($config->{'TagInactive'}));
	$table->MakeValid($inactive, $cgi->submit($config->{'TxtActive'}), 'middle');

	# Bottom of page
	$table->MakeBottom();

	# Embed
	print $cgi->end_form(), "\n";
}


#####################
# Activity Handler
#
# Will handle the activation de-activation of puppets as designated by the user
# this is to reduce the long list of puppets for selection when some are rarely
# used if ever.

sub ActivityHandler
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $config   = shift;					# Configuration tags	

	my $act      = shift;					# Active list
	my $sus      = shift;					# Suspended list

	my $active;						# Activity flag
	my $puppet;						# Puppet to handle

	my @puppets;						# Puppets to handle


	#####################
	# Program area	

	# Find appropraite values
	$active  = (defined($param->{$config->{'TxtActive'}})) ? 'NULL'  : 'LOCALTIMESTAMP'; 
	@puppets = (defined($param->{$config->{'TxtActive'}})) ? @{$sus} : @{$act};

	# Update entries
	eval
	{
		# Loop
		foreach $puppet (@puppets)
		{
			# Update activity
			$database->Write("UPDATE Puppet
				SET PuppetActive=$active
				WHERE PuppetName=?
				  AND PuppeteerLogin=?",

			 $puppet,
			 $param->{'USER'}
			);
		}

		# Commit changes
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}

	# Cleanup
	$param->Cleanup('PUPPET');
}

#####################
# Image Legal
#
# Determines if an image is actually legal based on a regular expression search
# through a list of known banned images.

sub ImageLegal
{
	####################
	# Data Members

	my $database = shift;					# Database handle	
	my $link     = shift;					# Link to image

	my $ban;						# Specific ban
	my @bans;						# List of bans


	####################
	# Program Area

	# Get list of banned images
	$database->GetListImageBan(\@bans);


	# Cycle through
	foreach $ban (@bans)
	{
		# Verify against link
		if ($link =~ /^$ban/si)
		{
			# Return illegal
			return 0;
		}		
	}


	# Return legal
	return 1;
}

#####################
# Image Size
#
# Downloads a picture off the net and determins it's size as to be able place 
# the defined sizes on the database

sub ImageSize
{
        ####################
        # Data Members

	my $database = shift;					# Database handle	
        my $link     = shift;					# URL to picture

        my $file;						# Name of file
	my $home;						# Homepage handler
	my $image;						# Image handle

	my $width;						# Image width
	my $height;						# Image height
	my $size;						# Image size

	my %system;						# System hash


        ####################
        # Program Area

	# Link to Hash
	$database->GetHashSystem(\%system);

	# Initiate Image and Homepage/File handles
	$image = new eRealms::Image($database, $system{'LocTemp'});
	$home  = new eRealms::Home($database, $system{'LocTemp'});


	# Fetch image
	($file) = $home->FileFetch($link);

	# Retreive demensions
	($width,
	 $height,
         $size) = $image->Size($file);

	# Unlink
	$home->FileUnlink($file); 


	# Return values
	return $width, $height, $size;
}


#####################
# InitialDisplay
#
# Handles the general output of the initial display screen.  Which will allow one
# to change password, E-Mail addresses and/or passwords.

sub InitialDisplay
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $audit;						# Auditing handle
	my $filter;						# Filter handle

	my $list;						# List of puppets
	my $conf;						# Confirmation
	my $newchar;						# New character
	my $clone;						# Cloning

	my $buttonmake;						# Create button
	my $buttonrem;						# Button remove
	my $buttonview;						# Button view

	my $iclone;						# Inline parameter clone
	my $iactvt;						# Inline parameter activity
	my $igroup;						# Inline parameter group
	my $url;						# Self-reference link

	my $tag;						# Tagline
	my $title;						# Title
	my $count;						# Active puppets

	my $choice;						# Choice of the matter
	my $message;						# Display message

	my @puppets;						# Array of puppets

	my %args;						# Default arguments


	#####################
	# Program area	
	
	# Create instance
	$audit    = new eRealms::Audit($database, $cgi, $param);
	$filter   = new eRealms::Filter();

	# Evaluation
	eval
	{
		# Addition operation
		if ( defined($param->{$config->{'TxtSubmitAddition'}}) ) 
		{
			# Strip HTML
			$param->{'PUPADD'} =  $filter->StripHTML($param->{'PUPADD'});

			# Additional filtering
			$param->{'PUPADD'} =~ s/~//gs;


			# Retreive information
			($choice) = $database->DataGet("SELECT PuppetName 
				FROM  Puppet 
				WHERE PuppetName=?
				  AND (PuppeteerLogin=?
				   OR PuppetPool='system')", 

			 $param->{'PUPADD'},
			 $param->{'USER'}
			);


			# Verification
			if ((length($param->{'PUPADD'}) > 2)
			 && ($param->{'PUPADD'} !~ /^\s/)
			 && (!defined($choice)))
			{
				# Tagline handling
				# Pull taglines
				if (defined($param->{'PUPCLONE'}))
				{
					# Pull from database
					($title, $tag) = $database->DataGetPuppetFormat(
						$param->{'PUPPET'}, 
						$param->{'USER'}
					 );
				}


				# Insertion upon acceptance
				$database->Write("INSERT INTO Puppet 
					(PuppetName, 
       	                          PuppeteerLogin,
					 PuppetTitle,
					 PuppetTag)
					VALUES(?,?,?,?)",

				 $param->{'PUPADD'}, 
				 $param->{'USER'},
				 $title,
				 $tag);

				# Record change
				$audit->PuppetAdd($param->{'PUPADD'});


				# Display status
				$message = $config->{'MsgCreate'};
			}
			else
			{ $message = $config->{'MsgExists'}; }
		}

		# Removal operation
		if ((defined($param->{$config->{'TxtSubmitDelete'}}))
		 && (defined($param->{'CONFIRM'})))
		{
			# Retreive information
			($choice) = $database->DataGet('SELECT PuppetName
				FROM Puppet 
				WHERE PuppetName=? 
				  AND PuppeteerLogin=?', 

			 $param->{'PUPPET'}, 
			 $param->{'USER'}
			);


			# Verification
			if (defined($choice))
			{
				# Removal of primary tables
				$database->Write("DELETE FROM Puppet 
					WHERE PuppetName=?
					  AND PuppeteerLogin=?", 

				 $param->{'PUPPET'},
				 $param->{'USER'}
				);

				# Record change
				$audit->PuppetRem($param->{'PUPPET'});


				# Message of success
				$message = $config->{'MsgDeletePass'}; 
			}		
			else
			{ $message = $config->{'MsgDeleteFail'}; }
		}

		# Commit
		$database->Commit();
	};

	# If block failed
	if ($@)
	{
		# Issue warning
		warn("Transaction aborted: $@");

		# Undo incomplete changes
		$database->Rollback();
	}


	# Cleanup
	$param->Cleanup('CONFIRM','PUPCLONE','PUPHTML');


	# Display of values
	# Retreive a list of puppets
	$database->GetListPuppetComplete(\@puppets, $param->{'USER'});

	# Total
	$count = @puppets;


	# Buttons
	# Default arguments
	# Define base arguments
	%args = (
		USER  => $param->{'USER'},
		CRYPT => $param->{'CRYPT'}
	 );

	# Generate Inline
	# Clone inline
	$iclone = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(%args, CLONE=>'True', ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(%args, CLONE=>'True');

	# Group inline
	$igroup = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(%args, GROUP=>'True', ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(%args, GROUP=>'True');

	# Activity inline	
	$iactvt = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(%args, ACTIVITY=>'True', ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(%args, ACTIVITY=>'True');

	# Pull URL
	$url    = $cgi->url(-absolute=>1);


	# Define message
	$message = (defined($message)) ? $message : '';


	# Prepare widgets
	$list    = $cgi->scrolling_list('PUPPET', \@puppets, $puppets[0], 10);
	$newchar = $cgi->textfield('PUPADD', '', 25, 30);

	# Checkboxes
	$conf    = $cgi->checkbox(-name=>'CONFIRM', -label=>'');
	$clone   = $cgi->checkbox(-name=>'PUPCLONE', -label=>'');

	# Buttons
	$buttonview = $cgi->submit($config->{'TxtSubmitView'});
	$buttonrem  = $cgi->submit($config->{'TxtSubmitDelete'});
	$buttonmake = $cgi->submit($config->{'TxtSubmitAddition'});


	# Selection/removal
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";


	# Search and replace template
	$tmpl->Show('TmplPupList',
                LACTVT => "HREF=\"$url$iactvt\"",
                LCLONE => "HREF=\"$url$iclone\"",
		LGROUP => "HREF=\"$url$igroup\"",
		MSTATS => $message,
		MCOUNT => $count,
		WCLNE  => $clone,
		WCONF  => $conf,
		WLIST  => $list,
		WNEWC  => $newchar,
		WDELE  => $buttonrem,
		WMAKE  => $buttonmake,
		WVIEW  => $buttonview
	 );


	# End form
	print $cgi->end_form();
}


#####################
# PuppetClone
#
# Allows the cloning (as in the sheep) of tagline and titles from one handle
# to others.  Basically a straight forward process and may speed up mass
# changes. 

sub PuppetClone
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $targets  = shift;					# Array of targets

	my $option   = new eRealms::Option();			# Option handle

	my $mode;						# Copy mode
	my $puppet;						# Puppets selection
	my $target;						# List of targets

	my $inline;						# Inline parameter handling
	my $url;						# Return link

	my $count;						# Counter
	my $tag;						# Tagline
	my $title;						# Title
	my $message;						# Display message
	my $placeholders;					# Placeholders

	my @modes;						# Clone modes
	my @puppets;						# Array of puppets


	#####################
	# Program area	

	# Clone handling
	# Only if necessary
	if ((defined($param->{$config->{'TxtClone'}}))
	 && (defined($param->{'PUPPET'})))
	{
		# Placeholders
		# Fetch total
		$count = @{$targets};	

		# Generate placeholders
		$placeholders = "?," x $count;
		chop($placeholders);


		# Pull tagline and tite
		($title, $tag) = $database->DataGetPuppetFormat(
			$param->{'PUPPET'}, 
			$param->{'USER'}
		 );

		# Discover mode
		$mode = $option->FindClone($config->{'OptClone'}, $param->{'MODE'});


		# Operate based on mode
		if ($count > 0)
		{
			# Tagline only
			if ($mode eq 'tag')
			{
				# Update of tagline only
				$database->Quick("UPDATE Puppet
					SET PuppetTag=?
					WHERE PuppeteerLogin=?
					  AND PuppetName IN ($placeholders)",

				 $tag,
				 $param->{'USER'},
				 @{$targets}
				);
			}

			# Title
			elsif ($mode eq 'title')
			{
				# Update of title only
				$database->Quick("UPDATE Puppet
					SET PuppetTitle=?
					WHERE PuppeteerLogin=?
					  AND PuppetName IN ($placeholders)", 

				 $title,
				 $param->{'USER'},
				 @{$targets}
				);
			}

			# Both title and tag
			elsif ($mode eq 'both')
			{
				# Update of tagline only
				$database->Quick("UPDATE Puppet
					SET PuppetTag=?,
					    PuppetTitle=?
					WHERE PuppeteerLogin=?
					  AND PuppetName IN ($placeholders)",

				 $tag,
				 $title,
				 $param->{'USER'},
				 @{$targets}
				);
			}

			# Warm and fuzzy message
			$message = $config->{'MsgClone'};
		}

		else
		{
			# Tap on wrist
			$message = $config->{'MsgCount'};
		}

		# Cleanup
		$param->Cleanup($config->{'TxtClone'});
	}

	
	# Pull valies
	# Retreive a list of puppets
	$database->GetList(\@puppets, "SELECT PuppetName 
		FROM Puppet 
		WHERE PuppeteerLogin=? 
		  AND PuppetActive IS NULL
		ORDER BY PuppetName", 

	 $param->{'USER'}
	);

	# Mode
	$option->Split(\@modes, $config->{'OptClone'});

	# Back button
	# Generate Inline and URL
	$inline = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'}, ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'});	
	$url    = $cgi->url(-absolute=>1);


	# Prepare widgets
	$mode   = $cgi->radio_group('MODE', \@modes, $modes[2], 'true');
	$puppet = $cgi->scrolling_list('PUPPET', \@puppets, $puppets[0], 5);
	$target = $cgi->scrolling_list('TARGET', \@puppets, $puppets[0], 5, 'true');


	# Information display
	# Title
	$tmpl->ShowMain($config->{'TagPupClone'});

	# Selection/removal
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";


	# Table top
	$table->MakeTop();

	# Back button
	$table->MakeSingle("<CENTER>" 
	  . $tmpl->Pass('TagTabbed', MLINK=>"$url$inline", MNAME=>$config->{TxtBack})
	  . "</CENTER>"
	);
	$table->MakeBlank();

	# Status message
	if (defined($message)) 
	{ 
		$table->MakeSingle($message); 
		$table->MakeBlank(); 
	}
	
	# Source and Target type
	$table->MakeSingle($tmpl->PassTitle($config->{'TagCloneTarget'}));
	$table->MakeTertiary($puppet, $cgi->submit($config->{'TxtClone'}), $target, 'middle');
	$table->MakeBlank();

	# Mode
	$table->MakeSingle($tmpl->PassSub($config->{'TagCloneMode'}));
	$table->MakeSingle($mode);
	$table->MakeBottom();

	# End form
	print $cgi->end_form();
}


#####################
# PuppetGroup
#
# Will permit users to group their puppets into different categories based
# on their needs.   This would only show up in realms themselves, as other
# places do not need that level of complexity.

sub PuppetGroup
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash
	my $are      = shift;					# Generated in group
	my $not      = shift;					# Generated lesser then group

	my $inline;						# Paranmeter list
	my $url;						# Self referencing link

	my $control;						# Control widget
	my $count;						# Number of groups
	my $group;						# Singular group
	my $puppet;						# Individual puppet

	my $ware;						# Generated in group
	my $wnot;						# Generated lesser then group

	my @are;						# Are part of the group
	my @not;						# Not part of the group
	my @groups;						# List of groups

	my %args;						# Argument list


	#####################
	# Program area	

	# Information handling
	# Creation
	if ((defined($param->{$config->{'TxtSubmitAddition'}}))
	 && (defined($param->{'NEWGROUP'}))
	 && ($param->{'NEWGROUP'} =~ /^\w{3,10}$/))
	{
		# Create new group
		$database->Quick("INSERT INTO Association
			(PuppeteerLogin,
			 AssociationName)
			VALUES(?,?)",

		 $param->{'USER'},
		 $param->{'NEWGROUP'}
		);

		# Change current group
		$param->{'CURRENT'} = $param->{'NEWGROUP'};

		# Cleanup
		$param->Cleanup('NEWGROUP');
	}

	# Removal
	elsif ((defined($param->{$config->{'TxtSubmitDelete'}}))
	  && (defined($param->{'CURRENT'})))
	{
		# Evaluate operation
		eval
		{
			# Remove from association
			$database->Write("DELETE FROM Association
				WHERE PuppeteerLogin=?
				  AND AssociationName=?",

			 $param->{'USER'},
			 $param->{'CURRENT'}
			);

			# Make sure no puppets carry this
			$database->Write("UPDATE Puppet
				SET PuppetAssociation=NULL
				WHERE PuppeteerLogin=?
				  AND PuppetAssociation=?",

			 $param->{'USER'},
			 $param->{'CURRENT'}
			);

			# Commit
			$database->Commit();

			# Cleanup
			$param->Cleanup('CURRENT');
		};

		# Operation failure
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}

	# Puppet updates
	# Additions
	elsif ((defined($param->{$config->{'TxtMoveIn'}}))
	  && (defined($param->{'CURRENT'})))
	{
		# Evaluate
		eval
		{
			# Loop through
			foreach $puppet (@{$not})
			{
				# Update puppets group
				$database->Write("UPDATE Puppet
					SET PuppetAssociation=?
					WHERE PuppeteerLogin=?
					  AND PuppetName=?",

				 $param->{'CURRENT'},
				 $param->{'USER'},
				 $puppet
				);
			}

			# Commit changes
			$database->Commit();
		};

		# Operation failure
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}
                
	# Removal
	elsif ((defined($param->{$config->{'TxtMoveOut'}}))
	  && (defined($param->{'CURRENT'})))
	{
		# Evaluate
		eval
		{
			# Loop through
			foreach $puppet (@{$are})
			{
				# Update puppets group
				$database->Write("UPDATE Puppet
					SET PuppetAssociation=NULL
					WHERE PuppeteerLogin=?
					  AND PuppetName=?",

				 $param->{'USER'},
				 $puppet
				);
			}

			# Commit changes
			$database->Commit();
		};

		# Operation failure
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");

			# Undo incomplete changes
			$database->Rollback();
		}
	}
                

	# Pull values
	# Retreive a list of puppets
	$database->GetListGroup(\@groups, $param->{'USER'});

	# Number of groups
	$count = @groups;


	# Default argument list
	%args = (
		USER  => $param->{'USER'},
		CRYPT => $param->{'CRYPT'}
	);

	# Add room if necessary
	$args{'ROOM'} = $param->{'ROOM'} if (defined($param->{'ROOM'}));


	# Back button
	# Generate Inline and URL
	$inline = $param->EmbedInline(%args);
	$url    = $cgi->url(-absolute=>1);

	# Information display
	# Title
	$tmpl->ShowMain($config->{'TagPupClone'});

	# Selection/removal
	# Start form
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat()), "\n";


	# This user is picky or has too many handles
	if ($count >= 1)
	{
		# Ensure definition
		$param->{'CURRENT'} = (defined($param->{'CURRENT'})) ? $param->{'CURRENT'} : $groups[1];

		# Pull lists
		# Not associated
		$database->GetList(\@not, "SELECT PuppetName
			FROM Puppet
			WHERE PuppeteerLogin=?
			  AND (PuppetAssociation <> ?
			   OR PuppetAssociation IS NULL)
			ORDER By PuppetName",

		 $param->{'USER'},
		 $param->{'CURRENT'}
		);
		
		# Associated
		$database->GetList(\@are, "SELECT PuppetName
			FROM Puppet
			WHERE PuppeteerLogin=?
			  AND PuppetAssociation=?
			ORDER BY PuppetName",

		 $param->{'USER'},
		 $param->{'CURRENT'}
		);
		
		# Generated widgets
		$ware = $cgi->scrolling_list('ARE', \@are, $are[0], 5, 'true');
		$wnot = $cgi->scrolling_list('NOT', \@not, $not[0], 5, 'true');

		# Control widget
		$control = $cgi->submit($config->{'TxtMoveIn'}) . "<BR><BR>" . $cgi->submit($config->{'TxtMoveOut'});


		# HTML display
		# Generate group buttons
		print "<CENTER>";

		# Display back button
		$tmpl->Show('TagTabbed', MLINK=>"$url$inline", MNAME=>$config->{TxtBack});


		# Necessary addition
		$args{'GROUP'} = 'True';

		# Add all groups in
		foreach $group (@groups)
		{
			# Generate group inline
			$inline = $param->EmbedInline(%args, CURRENT=>$group);

			# Display new link
			$tmpl->Show('TagTabbed', MLINK=>"$url$inline", MNAME=>$group);
		}
	
		# Close field
		print "</CENTER>\n<BR>\n";


		# Table top
		$table->MakeTop();

		# Primary control
		$table->MakeSingle($tmpl->PassTitle($config->{'TagGroupMove'} . $param->{'CURRENT'}));
		$table->MakeTertiary($wnot, $control, $ware, 'middle');
		$table->MakeBlank();

		# Removal Option
		$table->MakeSingle($tmpl->PassTitle($config->{'TagGroupDel'}));
		$table->MakeValid(
		 	$tmpl->PassSub($param->{'CURRENT'}),	
			$cgi->submit($config->{'TxtSubmitDelete'})
		);
		$table->MakeBlank();
	}

	# Nothing to display
	else
	{
		# HTML widgets
		# Simply generate back button
		print "<CENTER>" . $tmpl->Pass('TagTabbed', MLINK=>"$url$inline", MNAME=>$config->{TxtBack}) . "</CENTER>\n<BR>\n";

		# Table top
		$table->MakeTop();
	}

	# Group creation is universal
	$table->MakeSingle($tmpl->PassTitle($config->{'TagGroupAddition'}));
	$table->MakeValid(
		$tmpl->PassSub($config->{'TagGroupName'}),
		$cgi->textfield('NEWGROUP', '', 10, 10)
	 );
	$table->MakeValid(
		'&nbsp;',
		$cgi->submit($config->{'TxtSubmitAddition'})
	 );
	$table->MakeBottom();

	# End form
	print $cgi->end_form();
}


#####################
# PuppetControl
#
# Sub-routine that will handle the change of information for a puppet.  This
# is assuming that the submit button was pressed.

sub PuppetControl
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter handle
	my $table    = shift;					# Table handle
	my $tmpl     = shift;					# Template handle
	my $config   = shift;					# Config hash

	my $statement;						# Query handle

	my $admin;						# Administrative access
	my $js;							# Javascript control
	my $mod;						# Moderator access
	my $super;						# Supervisory access
	my $ver;						# Verified user

	my $choice   = 0;					# Choice of the matter
	my $info;						# Puppet information
	my $gummies;						# Gummies html
	my $gummy;						# Default entry
	my $printout;						# Printout option

	my $inline;						# Parameters
	my $url      = $cgi->url(-absolute=>1);			# Self reference

	my @gummies;						# List of gummies


	#####################
	# Program area

	# Change of information and verify for appripriate button press
	if ( defined($param->{$config->{'TxtSubmitButton'}}) )
	{ 
		PuppetControlWrite($database, $param, $config); 
	}

	# Prinout handle
	$printout = $param->{'PRINTOUT'};
	$param->Cleanup('PRINTOUT');


	# Data Retreival
	# Determin privileges
	($super, $admin, $mod, $ver) = $database->DataGetPrivs($param->{'USER'});
	($js) = $database->DataGetJavascript($param->{'USER'});


	# Back button
	# Generate Inline
	$inline = (defined($param->{'ROOM'}))
	  ? $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'}, ROOM=>$param->{'ROOM'})
	  : $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'});	


	# Gather information
	# Prepare and execute
	$database->Pull(\$statement, "SELECT
			 PuppetName       AS \"PuppetName\",
			 PuppetActive     AS \"PuppetActive\",
			 PuppetBio        AS \"PuppetBio\",
			 PuppetDefault    AS \"PuppetDefault\",
			 PuppetDisplay    AS \"PuppetDisplay\",
			 PuppetGummy      AS \"PuppetGummy\",
			 PuppetPic        AS \"PuppetPic\",
			 PuppetPicHeight  AS \"PuppetPicHeight\",
			 PuppetPicSize    AS \"PuppetPicSize\",
			 PuppetPicWidth   AS \"PuppetPicWidth\",
			 PuppetPool       AS \"PuppetPool\",
			 PuppetTag        AS \"PuppetTag\",
			 PuppetTextColour AS \"PuppetTextColour\",
			 PuppetTextSize   AS \"PuppetTextSize\",
			 PuppetTitle      AS \"PuppetTitle\"
			FROM  Puppet
			WHERE PuppetName=? 
			  AND PuppeteerLogin=?",

	 $param->{'PUPPET'},
	 $param->{'USER'}
	);

	# Retreive into hash
	$info = $statement->fetchrow_hashref();


	# Gummy list
	# Retreive icons general or canon
	if ($info->{'PuppetPool'} eq 'system')
	{
		# Default and general
		GetGummyDefault($database, \@gummies);
		GetGummyGeneral($database, \@gummies);
	}
	else
	{
		# Simply canon handles
		GetGummyCanon($database, \@gummies);
	}

	# Administrative, Supservisory and Moderator gummies
	GetGummyContrib($database, \@gummies) if ($ver   eq 'yes');
	GetGummyAdmin($database, \@gummies)   if ($admin eq 'yes');
	GetGummySuper($database, \@gummies)   if ($super eq 'yes');
	GetGummyMod($database, \@gummies)     if ($mod   eq 'yes');

	# None option
	push(@gummies, $config->{'TxtNone'})  if ($ver   eq 'yes');


	# Reverse lookup gummy
	($gummy)  = ($info->{'PuppetGummy'} ne 'NIL')
	  ? $database->DataGetGummyName($info->{'PuppetGummy'})
	  : $config->{'TxtNone'};

	# Determin default and format
	$gummies  = $cgi->popup_menu('GUMMY', \@gummies, $gummy);


	# Add hint for users
	$info->{'PuppetPic'} = (defined($info->{'PuppetPic'})) ? $info->{'PuppetPic'} : 'http://';
	$info->{'PuppetBio'} = (defined($info->{'PuppetBio'})) ? $info->{'PuppetBio'} : 'http://';


	# Display of information
	# Javascript
	print "$config->{SetFormat}\n\n" if ($js eq 'yes');


	# Title
	$tmpl->ShowMain($config->{'TagPupDetails'});

	# Star form
	print $cgi->start_form(-name=>'PUPPET');
	print $param->EmbedNormal($param->Flat(), PUPPET=>$param->{'PUPPET'}, POOL=>$info->{'PuppetPool'}), "\n";

	# Table top
	$table->MakeTop();

	# Back button
	$table->MakeSingle("<CENTER><A HREF=\"$url$inline\">$config->{TagBack}</A></CENTER>");
	$table->MakeBlank();

	# Most clueful information
	$table->MakeSingle($tmpl->PassTitle($info->{'PuppetName'}));
	$table->MakeBlank();

	# Image information
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImageLink'}),
	 $cgi->textfield('PUPIMG', $info->{'PuppetPic'}, 30, 200));
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImgWidth'}),
	 "<I>$info->{'PuppetPicWidth'}</I>");
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImgHeight'}),
	 "<I>$info->{'PuppetPicHeight'}</I>");
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupImgSize'}),
	 "<I>$info->{'PuppetPicSize'}</I>");
	$table->MakeBlank();

	# Bio Url
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupBio'}),
	 $cgi->textfield('PUPBIO', $info->{'PuppetBio'}, 30, 200));
	$table->MakeValid(
	 $tmpl->PassSub($config->{'TagPupGummy'}),
	 $gummies);
	$table->MakeBlank();

	# Checkboxes
	# Active Puppets
	unless (defined($info->{'PuppetActive'}))
	{ 
		# Active
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagPupActive'}), 
		 $cgi->checkbox(-name=>'PUPACT', -checked=>'on', -label=>'')); 
	}
	else
	{ 
		# Inactive
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagPupActive'}), 
		 $cgi->checkbox(-name=>'PUPACT', -label=>'')); 
	}

	# Default Puppets
	if ($info->{'PuppetDefault'} eq 'yes')
	{ 	
		# Active
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagPupDefault'}), 
		 $cgi->checkbox(-name=>'PUPDEF', -checked=>'on', -label=>'')); 
	}
	else
	{ 
		# Inactive
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagPupDefault'}), 
		 $cgi->checkbox(-name=>'PUPDEF', -label=>'')); 
	}


	# Display Puppeteer information
	if ($info->{'PuppetDisplay'} eq 'yes')
	{ 	
		# Active
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagPupDisplay'}), 
		 $cgi->checkbox(-name=>'PUPDIS', -checked=>'on', -label=>'')); 
	}
	else
	{ 
		# Inactive
		$table->MakeValid(
		 $tmpl->PassSub($config->{'TagPupDisplay'}), 
		 $cgi->checkbox(-name=>'PUPDIS', -label=>'')); 
	}

	# Spacer
	$table->MakeBlank();


	# Display to realm
	if (defined($param->{'ROOM'}))
	{
		# Display Puppeteer information
		if (defined($printout))
		{ 	
			# Active
			$table->MakeValid(
			 $tmpl->PassSub($config->{'TagPupPrintout'}), 
			 $cgi->checkbox(-name=>'PRINTOUT', -checked=>'on', -label=>'')); 
		}
		else
		{ 
			# Inactive
			$table->MakeValid(
			 $tmpl->PassSub($config->{'TagPupPrintout'}), 
			 $cgi->checkbox(-name=>'PRINTOUT', -label=>'')); 
		}
	}

	# Spacer
	$table->MakeBlank();


	# More complexe title
	# Header
	if ($js ne 'yes')
	{
		# Basic look and feel
		$table->MakeSingle($tmpl->PassTitle($config->{'TagPupTitle'}));
	}
	else
	{
		# Advanced look
		$table->MakeSingle($tmpl->Pass('TmplFormat', 
			MFUNC => 'strTitle',
			MNAME => $config->{'TagPupTitle'}
		 )); 
	}

	# Widget
	$table->MakeSingle($cgi->textarea(
		-name=>'PUPTIT',
		-rows=>3,
		-columns=>65,
		-wrap=>'SOFT',
		-default=>$info->{'PuppetTitle'})
	 );
	$table->MakeBlank();


	# More complexe tagline
	# Footer
	if ($js ne 'yes')
	{
		# Basic look and feel
		$table->MakeSingle($tmpl->PassTitle($config->{'TagPupTagline'}));
	}
	else
	{
		# Advanced look
		$table->MakeSingle($tmpl->Pass('TmplFormat',
			MFUNC => 'strTagline',
			MNAME => $config->{'TagPupTagline'}
		 )); 
	}

	# Physical widget
	$table->MakeSingle($cgi->textarea(
		-name=>'PUPTAG',
		-rows=>3,
		-columns=>65,
		-wrap=>'SOFT',
		-default=>$info->{'PuppetTag'})
	 );
	$table->MakeBlank();


	# Finish table
	$table->MakeValid($cgi->submit($config->{'TxtSubmitButton'}), $cgi->reset($config->{'TxtResetButton'}));
	$table->MakeBottom();


	# End of form
	print $cgi->end_form();

	# Finish statement
	$statement->finish();
}


#####################
# PuppetControlWrite
#
# Sub-routine that will handle the change of information for a puppet.  This
# is used as a sub process to handle database interaction.

sub PuppetControlWrite
{
	#####################
	# Data members
	my $database = shift;					# Database handle
	my $param    = shift;					# Parameter handle
	my $config   = shift;					# Config hash

	my $comm;						# Post handler
	my $filter   = new eRealms::Filter();			# Filter handle

	my $legal;						# Image legal
	my $height;						# Image height
	my $size;						# Image size
	my $width;						# Image width

	my $cbio;						# Bio
	my $cgum;						# Gummy
	my $ctag;						# Cleaned tag
	my $ctit;						# Cleaned title

	my $active;						# Active flag
	my $default;						# Default toggle
	my $display;						# Display of information
	my $file;						# Filename
	my $name;						# Name of gummy

	my $lang;						# Language filter
	my @lang;						# Array of banned words


	#####################
	# Program area

	# Determine active state
	($active) = $database->DataGetPuppetActive($param->{'PUPPET'}, $param->{'USER'});


	# Image handling
	# If valid URL
	if ((length($param->{'PUPIMG'}) > 10)
	 && ($param->{'PUPIMG'} !~ /^http:\/\/\d{1,3}\.\d{1,3}\.\d{1,3}/i)) 
	{
		# If legal
		if (ImageLegal($database, $param->{'PUPIMG'}))
		{	
			# Capture and determin size
			($width, $height, $size, $_) = ImageSize($database, $param->{'PUPIMG'}); 

			# Make sure image is valid
			$param->{'PUPIMG'} = defined($width) ? $param->{'PUPIMG'} : undef;

	
			# Update database with new information
			$database->Quick("UPDATE Puppet
				SET PuppetPic=?, 
				    PuppetPicWidth=?, 
				    PuppetPicHeight=?,
				    PuppetPicSize=?
				WHERE PuppetName=? 
				  AND PuppeteerLogin=?", 

			 $param->{'PUPIMG'}, 
			 $width, 
			 $height, 
			 $size,
			 $param->{'PUPPET'}, 
			 $param->{'USER'}
			); 
		}
	}
	else
	{
		# Update database with nulls
		$database->Quick("UPDATE Puppet 
			SET PuppetPic=NULL, 
		            PuppetPicWidth=NULL, 
			    PuppetPicHeight=NULL,
			    PuppetPicSize=NULL
			WHERE PuppetName=? 
			  AND PuppeteerLogin=?", 

		 $param->{'PUPPET'}, 
		 $param->{'USER'}
		); 
	}


	# Assign default
	if (defined($param->{'PUPDEF'}))
	{
		# Database updates
		# Pull any stray tags
		$database->Quick("UPDATE Puppet
			SET PuppetDefault='no'
			WHERE PuppeteerLogin=?
			  AND PuppetDefault='yes'",

		 $param->{'USER'}
		);

		# Assign proper default
		$database->Quick("UPDATE Puppet
			SET PuppetDefault='yes'
			WHERE PuppetName=?
			  AND PuppeteerLogin=?
			  AND PuppetPool='system'",

		 $param->{'PUPPET'},
		 $param->{'USER'}
		);
	}


	# Active puppets
	# Activate
	if ((defined($param->{'PUPACT'})) 
	 && (defined($active)))
	{
		# Remove timestamp
		$database->Quick("UPDATE Puppet
			SET PuppetActive=NULL
			WHERE PuppetName=?
			  AND PuppeteerLogin=?",

		 $param->{'PUPPET'},
		 $param->{'USER'}
		);
	}

	# Deactivate
	elsif ((!defined($param->{'PUPACT'}))
	 && (!defined($active)))
	{
		# Remove timestamp
		$database->Quick("UPDATE Puppet
			SET PuppetActive=LOCALTIMESTAMP
			WHERE PuppetName=?
			  AND PuppeteerLogin=?",

		 $param->{'PUPPET'},
		 $param->{'USER'}
		);
	}


	# Data handling
	# Truncate and clean output
	$ctag = substr($param->{'PUPTAG'}, 0, 1024);
	$ctit = substr($param->{'PUPTIT'}, 0, 1024);
	$cbio = $param->{'PUPBIO'};

	# Pull value
	$display = $param->{'PUPDIS'};


	# Gummy handling
	($cgum) = ($param->{'GUMMY'} ne $config->{'TxtNone'})
	  ? $database->DataGetGummyPath($param->{'GUMMY'})
	  : 'NIL';	


	# Tag handling
	($ctag) = $filter->SafeHTML($ctag);
	($ctit) = $filter->SafeHTML($ctit);

	# Language filter
	# Retreive all words
	$database->GetList(\@lang, "SELECT RestrictionValue
		FROM  Restriction
		WHERE RestrictionType='word'
		  AND RealmName='system'
		GROUP BY RestrictionValue");

	# Language filtering
	while ($lang = shift(@lang))
	{
		# Avoid host stupidity
		$lang = quotemeta($lang);

		# Run a global replace
		$ctag =~ s/$lang/****/gsi;
		$ctit =~ s/$lang/****/gsi;
	}


	# Database update
	# Tagline, Title and Bio link
 	$ctag = (length($ctag) > 3)  ? $ctag : undef;
 	$ctit = (length($ctit) > 3)  ? $ctit : undef;
 	$cbio = ((length($cbio) > 12) && ($cbio =~ /^http:\/\//is)) ? $cbio : undef; 

	# Information display
	$display = (defined($display)) ? 'yes' : 'no';

	# Update fields
	$database->Quick("UPDATE Puppet 
		SET PuppetTag=?, 
		    PuppetTitle=?,
		    PuppetBio=?,
		    PuppetGummy=?,
		    PuppetDisplay=? 
		WHERE PuppetName=? 
		  AND PuppeteerLogin=?",

	 $ctag, 
	 $ctit, 
	 $cbio,
	 $cgum, 
	 $display,
	 $param->{'PUPPET'}, 
	 $param->{'USER'}
	);



	# Display
	if (defined($param->{'PRINTOUT'}))
	{
		# Define instance
		$comm  = new eRealms::Comm($database, $cgi);

		# Pull values
		my ($tsize, $tcolor, $tface) = $database->DataGetFont($param->{'PUPPET'}, $param->{'USER'});


		# Basic font setup
		$param->{'SIZE'}  = $tsize;
		$param->{'COLOR'} = $tcolor;
		$param->{'FACE'}  = $tface;

		# Identifying marks
		$param->{'CHAR'}    = $param->{'PUPPET'};
		$param->{'SOURCE'}  = $param->{'USER'}; 
		$param->{'TARGET'}  = $param->{'USER'}; 
		$param->{'OTARGET'} = $param->{'PUPPET'}; 

		# Set post
		$param->{'POST'}   = $config->{'MsgPupPrintout'};

		# Send post
		$comm->DataWrite($param);
	}

	# Cleanup
	$param->Cleanup(
	  'CHAR',
	  'COLOR',
	  'FACE',
	  'OTARGET',
	  'POOL',
	  'POST',
	  'PUPACT',
	  'PUPDEF',
	  'PUPDIS',
	  'SOURCE',
	  'TARGET'
	 );
}


#####################
# Puppet Gummy
#
# This function is a variant of the GetList function used to pull
# fields from the database.  However, since this one will not
# clean up the array which is generally desired. 

sub PuppetGummy
{
	#####################
	# Data members
	my $database = shift;					# Database handle

	my $statement;						# Database statement

	my $array    = shift;					# Reference to array
        my $class    = shift;					# Identification class                             
	my $realm    = shift;					# Realm name


	#####################
	# Program Area

	# Define realm
	$realm = (defined($realm)) ? $realm : 'system';


	# Prepare and execute
	$database->Pull(\$statement, "SELECT GummyName 
		FROM Gummy 
		WHERE GummyClass=?
		  AND RealmName=?
		ORDER BY GummyIdent, 
			 GummyName",

	 $class,
	 $realm
	);


	# Retrieve and place in array
	while ($_ = $statement->fetchrow())
	{ push(@{$array}, $_); }

	# End statement
	$statement->finish;
}

# Alias for administrators
sub GetGummyAdmin
{ PuppetGummy(shift, shift, 'admin'); }

# Alias for default gummy
sub GetGummyDefault
{ PuppetGummy(shift, shift, 'default'); }

# Alias for canon handles
sub GetGummyCanon
{ PuppetGummy(shift, shift, 'canon'); }

# Alias for contributor gummies
sub GetGummyContrib
{ PuppetGummy(shift, shift, 'contributor'); }

# Alias for general
sub GetGummyGeneral
{ PuppetGummy(shift, shift, 'general'); }

# Alias for moderators
sub GetGummyMod
{ PuppetGummy(shift, shift, 'mod'); }

# Alias for supervisors
sub GetGummySuper
{ PuppetGummy(shift, shift, 'super'); }
